#include "Log.h"
#include <stdio.h>
#include <string.h>
#include "simulator.h"

int received = 0;

Log::Log()
	{
	for(int i=0; i<numNodes; i++)
		for(int j=0; j<numNodes; j++)
			flows[i][j][0] = flows[i][j][1] = 0;

	ConsecutiveBins = 61;
	}

void Log::order()
	{
    ofstream tgs;
    tgs.open("..\\minibtma.tgs");

	list<logEntry>::iterator ei = --logList.end();

	while( ei != (logList.begin()))
		{
        tgs << ei->tx << ' ' << ei->rx << ' ' << ei->exid << ' ' << ei->sent << endl;

		ei--;
		}

	tgs.close();        
    }

void Log::logSuccess(Event e)
	{
	// Check that this exchange hasn't been logged already
	list<logEntry>::iterator ei = logList.begin();

	while( ei != (logList.end()))
		{
		if(e.getExid() == ei->exid)
			{
#ifdef DEBUG
	Simulator::foo << " Already Logged!";
	cout << " Already Logged!";
#endif
			return;
			}
		else
			ei++;
		}

	// log entry
	logEntry entry(e);
	logList.push_front(entry);

	// recalculate average delay
	int f = flows[e.getRx()][e.getTx()][0];
	flows[e.getRx()][e.getTx()][1] = (double)flows[e.getRx()][e.getTx()][1]*f/(f+1)+entry.delay/(f+1);

	// increment number of exchanges
	flows[e.getRx()][e.getTx()][0]++;

#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << " Entry Logged!";
    }
#endif
	}

void Log::show()
	{
	char s[25];
	int sum = 0;

	// calculate and display total number of successful exchanges
	for(int i=0; i<numNodes; i++)
		for(int j=0; j<numNodes; j++)
			sum += flows[i][j][0];

	Simulator::foo << "==================================================================" << endl
        << "  RESULTS:" << endl << endl
		<< sum << " Successful Exchanges:" << endl << endl;

	cout << "==================================================================" << endl
        << "  RESULTS:" << endl << endl
		<< " Successful Exchanges:" << sum << endl
      	<< "Received: " << received << endl << endl << endl;

	Simulator::foo << "      |Rx:            0";
	cout << "      |Rx:            0";
	for(int i=1; i<numNodes; i++)
		{
		sprintf(s,"%16d",i);
		Simulator::foo << s;
		cout << s;
		}
	Simulator::foo << endl;
	cout << endl;

	Simulator::foo << "------+";
	cout << "------+";
	for(int i=0; i<numNodes; i++)
		{
		Simulator::foo << "----------------";
		cout << "----------------";
		}
	Simulator::foo << endl;
	cout << endl;

	for(int i=0; i<numNodes; i++)
		{
		if(i==0)
			{
			Simulator::foo << "Tx: 0 |";
			cout << "Tx: 0 |";
			}
		else
			{
			sprintf(s,"%5d",i);
			Simulator::foo << s << " |";
			cout << s << " |";
			}

		for(int j=0; j<numNodes; j++)
			{
			sprintf(s,"%d(%d)",flows[j][i][0],flows[j][i][1]);
			Simulator::foo.width(16);
			Simulator::foo << s;
			cout.width(16);
			cout << s;
			}
		Simulator::foo << endl;
		cout << endl;
		}

	if(numNodes>12)
    	{
	    cout << endl;
    	long total = 0;
	    for(int i=0; i<numNodes; i++)
    		for(int j=0; j<numNodes; j++)
        		{
            	if(flows[i][j][0])
            		{
	                cout << "Tx: " << j << " Rx: " << i << " Packets Sent: " << flows[i][j][0] << endl;
                    Simulator::foo << "Tx: " << j << " Rx: " << i << " Packets Sent: " << flows[i][j][0] << endl;
    	            total += flows[i][j][0];
        	        }
            	}

	    cout << "Total Sent: " << total << endl;
        Simulator::foo << "Total Sent: " << total << endl;
        }
	}

void Log::icons(int i)
	{
    if(i>=6 && i<20)
    	i=6;
	else if(i>=20 && i<60)
    	i=7;
    else if(i>=60)
    	i=8;

    cons[i]++;
    }

void Log::consecutive()
	{
    for(int i=0; i<ConsecutiveBins; i++)
    	cons[i] = 0;

	int ctx[numNodes];

    for(int i=0; i<numNodes; i++)
    	ctx[i] = 0;

	list<logEntry>::iterator ei = logList.begin();

#ifdef n4f2
	while(ei != logList.end())
		{
        if(ei->tx == 0)
        	{
            icons(ctx[2]);
            ctx[2] = 0;
            }
        else if(ei->tx == 2)
        	{
            icons(ctx[0]);
            ctx[0] = 0;
            }

		ctx[ei->tx]++;
        ei++;
		}
#endif
#ifdef n6f3
	while(ei != logList.end())
		{
        if(ei->tx == 0)
        	{
            icons(ctx[2]);
            ctx[2] = 0;
            }
        else if(ei->tx == 2)
        	{
            icons(ctx[0]);
            ctx[0] = 0;
            icons(ctx[4]);
            ctx[4] = 0;
            }
		else if(ei->tx == 4)
        	{
            icons(ctx[2]);
            ctx[2] = 0;
            }

		ctx[ei->tx]++;
        ei++;
		}
#endif
#ifdef n10f5
	while(ei != logList.end())
		{
        if(ei->tx == 0)
        	{
            icons(ctx[2]);
            ctx[2] = 0;
            }
        else if(ei->tx == 2)
        	{
            icons(ctx[0]);
            ctx[0] = 0;
            icons(ctx[4]);
            ctx[4] = 0;
            }
		else if(ei->tx == 4)
        	{
            icons(ctx[2]);
            ctx[2] = 0;
            icons(ctx[6]);
            ctx[6] = 0;
            icons(ctx[8]);
            ctx[8] = 0;
            }
		else if(ei->tx == 6)
        	{
            icons(ctx[4]);
            ctx[4] = 0;
            }
        else if(ei->tx == 8)
        	{
            icons(ctx[4]);
            ctx[4] = 0;
            }

		ctx[ei->tx]++;
        ei++;
		}
#endif

    for(int i=0; i<numNodes; i++)
            icons(ctx[i]);

//    Simulator::fooc << endl;
//    for(int i=0; i < ConsecutiveBins; i++)
//    	Simulator::fooc << i << " " << cons[i] << endl;
    }
