#include <stdlib.h>
#include <conio.h>
#include "node.h"
#include "simulator.h"
#include "EventList.h"
#include "seeds.h"

bool data_received = false;
extern int received;

extern int adjMatrix[numNodes][numNodes];

extern bool isTx[numNodes];
extern int  recv[numNodes];
extern bool isRx[numNodes];
extern int  trns[numNodes];

int node::id_counter = 0;			// node id counter
Log* node::log = 0;
EventList* node::eventList;
Simulator* node::sim;

node::node()
	{
	id = id_counter++;
	waitingCTS = numTrans = 0;
    waitingDATA = false;
    rBT = 0;
    dataW = Event();
    
	for(int i=0; i<numNodes; i++)
		trans[i] = false;
	}

node::~node()
	{
    if(tgsfile.is_open())
    	tgsfile.close();

    id_counter--;
    }

void node::scheduleFirst()
	{
	if(isTx[id])
    	{
		char file[] = "..\\Txii.tgs";
    	file[5] = id/10+'0';
        file[6] = id%10+'0';
	    tgsfile.open(file);

        if(!tgsfile.is_open())
        	{
            cout << "ERROR:  Can't open " << file << endl;
            cout << isTx[id] << endl;
            getch();
            exit(0);
            }

        scheduleNew(sTime,Event());        
        }
	}

// Schedule New Transmission
void node::scheduleNew(long now, Event e)
	{
	long time;				// time of event
	long exid;				// exchange id
	int  tx;
	int  rx;
	int  dur;

    if(e.getType()==NULL)
    	{
	    tgsfile >> tx >> rx >> exid >> time >> dur;

		// Schedule the next exchange to the time it was generated or now, whichever is greater
		time = now > time? now : time ;

	    // Change time to match begining of next mini slot
    	time += MINISLOTTIME - (time%MINISLOTTIME);
	    dur  += MINISLOTTIME - (dur%MINISLOTTIME) -1;
	    Event ev(tx, rx, RTSOn, time+10, exid, dur, time);
		eventList->addEvent(ev);
        }
    else
    	{
        time = e.getGen();
        exid = e.getExid();
        tx = e.getTx();
        rx = e.getRx();
        dur = e.getRx();
	    Event ev(tx, rx, RTSOn, now+11, exid, dur, time);
		eventList->addEvent(ev);
        }
	}

// notify() is called by the simulator to indicate it is time for the event
void node::notify(Event e)
	{
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << "seeds: ";

    for(int i=0; i<numNodes; i++)
    	{
        if(isTx[i])
        	{
            Simulator::foo << i << ':' << ranks.seedOf(i) << ", ";
            }
        }

    Simulator::foo << endl;
    }
#endif

	if(e.getType()==RTSOn)							// Time to send RTS
		{
        int slot = ranks.whenIsMySlot(id,e.getRx());
		if( slot != 0 )
        	{
            long seedsResc = slot * MINISLOTTIME + e.getTime();

        	e.reschedule(seedsResc);
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << "RESCHEDULE! (seeds): " << slot << endl;
    }
#endif

			eventList->addEvent(e);
			return;
			}
		else if(rBT)					// Busy tone detected
        	{
			e.reschedule( nextSlotTime(e.getTime())+10);// reschedule
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << "RESCHEDULE! (Busy Tone)" << endl;
    }
#endif
			eventList->addEvent(e);
			return;
            }

		// Send RTS
		sim->broadcastEvent(id, e);

		// Schedule RTSOff
		Event off(id,e.getRx(),RTSOff, e.getTime()+RTSDuration,e.getExid(),e.getDur(),e.getGen());
		eventList->addEvent(off);
		}
	else if(e.getType()==RTSOff)
		{
		// Send RTS Off
		sim->broadcastEvent(id, e);
		waitingCTS = e.getExid();

		// Schedule NoReplyToRTS
		Event no(id, e.getRx(), NoReplyToRTS, nextSlotTime(e.getTime())-1, e.getExid(),e.getDur(),e.getGen());
		eventList->addEvent(no);

        dataW = Event(id, e.getRx(), DATAOn, nextSlotTime(e.getTime()), e.getExid(),e.getDur(),e.getGen());
		}
	else if(e.getType()==NoReplyToRTS)
		{
		if(waitingCTS == e.getExid())			// If waiting for CTS
			{
			// Schedule RTS
			Event rts(id,e.getRx(),RTSOn,nextSlotTime(e.getTime())+10,e.getExid(),e.getDur(),e.getGen());
			eventList->addEvent(rts);
			waitingCTS = 0;
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << "RESCHEDULE! (No Reply To RTS)" << endl;
    }
#endif
			}
#ifdef DEBUG
		else
        	{
if(Simulator::active)
	{
	Simulator::foo << "Ignored!" << endl;
    }
    		}
#endif
		}
	else if(e.getType()==CTSOn) // This is actually a BTr On
		{
        sim->broadcastBTr(id,true);

        if(rBT==1)
        	{
#ifdef DEBUG
if(Simulator::active)
	{
//	Simulator::foo << endl << "NRA sent!!!" << endl << endl;
    }
#endif
#ifdef NRACTS
			sim->broadcastEvent(id, e);

			Event ctsoff(id, e.getRx(), CTSOff, e.getTime()+CTSDuration, e.getExid(),e.getDur(),e.getGen());
			eventList->addEvent(ctsoff);
#endif
            }

        waitingDATA = true;

        // Schedule no reply to busy tone
        Event noreply(e.getTx(),e.getRx(),NoReplyToBTrOn,nextSlotTime(e.getTime())+DATATimeOut,0,0,0);
        eventList->addEvent(noreply);
		}
	else if(e.getType()==CTSOff)
		{
		// Send CTS Off
		sim->broadcastEvent(id, e);
		}
    else if(e.getType()==NoReplyToBTrOn)
		{
        if(waitingDATA)
        	{
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << " DATA NOT RECEIVED. BT Rx Off! " << endl;
    }
#endif
            waitingDATA = false;
            sim->broadcastBTr(id,false);
            }
		else
			{
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << " Ignored! " << endl;
    }
#endif
			}
        }
	else if(e.getType()==DATAOn)
		{
		sim->broadcastEvent(id, e);

		Event dataoff(id, e.getRx(), DATAOff,e.getTime()+e.getDur(),e.getExid(),e.getDur(),e.getGen());
		eventList->addEvent(dataoff);
		}
	else if(e.getType()==DATAOff)
		{
		// Send DATA Off
		sim->broadcastEvent(id, e);

        if(data_received)
        	{
            received++;
            scheduleNew(e.getTime(),Event());
            data_received = false;
            }
		else
			scheduleNew(e.getTime(),e);

        ranks.setRA(id,0);

        if(NRAGENIE)
			sim->NRA_ALL(id);
		}
	}

void node::receive(Event e)
	{
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << " Received by " << id << ": ";
    }
#endif

	// update info on neighbouring transmissions
	if(e.getType() == RTSOn || e.getType() == CTSOn || e.getType() == DATAOn || e.getType() == ACKOn)		
		{
		trans[e.getTx()] = true;
		numTrans++;

		if(numTrans==1)										// If one transmission is going on
			Ongoing = e;									//  keep track of this transmission
		else
        	{
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << "  Packet Collision!";
    }
#endif
			Ongoing = Event();								// Otherwise tx is corrupt
			}
		}
	else if(e.getType() == RTSOff || e.getType() == CTSOff || e.getType() == DATAOff
		|| e.getType() == ACKOff)
		{
		// If a packet was received correctly (there were no collisions during transmission)
		// process it here.
		if(Ongoing.getType()==null  || e.getTx() == id)
			{
			// Packet was destroyed by interference - do nothing
			}
		else if(e.getType()==RTSOff)
			{
			if(e.getRx() == id)						// If the RTS was sent to this node
				{
				// Schedule CTS
				Event cts(id, e.getTx(), CTSOn, e.getTime()+IFS, e.getExid(),e.getDur(),e.getGen());
				eventList->addEvent(cts);
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << "  BTr On Scheduled!";
    }
#endif
				}

#ifdef NRARTS
                ranks.setRA(e.getTx(),e.getDur()/MINISLOTTIME +1);
#endif
			}
        else if(e.getType()==CTSOff)
        	{
            if(e.getRx()!=id)
            	{
                ranks.setRA(e.getRx(),e.getDur()/MINISLOTTIME +1);
                }
            }
		else if(e.getType()==DATAOff)
			{
			if(e.getRx() == id)					// If the DATA was sent to this node
				{
				waitingDATA = false;
                sim->broadcastBTr(id,false);
                data_received = true;                

				// Log Success
				log->logSuccess(e);
				}
			}

		trans[e.getTx()] = false;							// update trans[]
		numTrans--;
		Ongoing = Event();
		}
#ifdef DEBUG
if(Simulator::active)
	{
	Simulator::foo << endl;
    }
#endif
	}

void node::rBusyTone(bool on)
	{
    if(on)
    	{
    	rBT++;

        if(waitingCTS)
        	{
            eventList->addEvent(dataW);
            waitingCTS = 0;
            }
        }
    else
    	{
    	rBT--;
        }

    }

