#include "seeds.h"
#include "simulator.h"
#include <stdlib.h>
#include <iostream.h>
#include "random\\Random.h"

extern int adjMatrix[numNodes][numNodes];

extern bool isTx[numNodes];
extern int  recv[numNodes];
extern bool isRx[numNodes];
extern int  trns[numNodes];

int nnn = (unsigned) time(0);  				// for random number generation
UniformStream urand(0,MAXSEED,nnn%1000);
int Seeds::current = 0;
int Seeds::seeds[numNodes][2*numNodes];

Seeds::Seeds()
	{
    for(int i=0; i<numNodes; i++)
    	{
        if(BACKOFFSCHEME == 1)
	        mods[i] = 0;
		else if(BACKOFFSCHEME == 2)
	        mods[i] = 1;
        }
    }

void Seeds::init()
	{
    for(int i=0; i<numNodes; i++)
    	{
    	for(int j=0; j<2*numNodes; j++)
    	    seeds[i][j] = urand();
		}
    }

void Seeds::regenerate_seeds()
	{
    if(++current == 2*numNodes)
    	{
        current = 0;

	    for(int i=0; i<numNodes; i++)
    		for(int j=0; j<2*numNodes; j++)
    		    seeds[i][j] = urand();
        }
    }

void Seeds::tick()
	{
    for(int i=0; i<numNodes; i++)
    	{
		if(BACKOFFSCHEME == 1)
        	{
            if(mods[i]>0)
		        mods[i]--;
            }
		else if(BACKOFFSCHEME == 2)
			{
	        mods[i] /= 3;

    	    if(mods[i] < 1)
        		mods[i] = 1;
            }
        }
    }

bool Seeds::first(int n1, int n2, int tn)
   	{
    int mod1=mods[n1], mod2=mods[n2];

    if(BACKOFFSCHEME == 1)
    	{
		int rank1, rank2;

   	    if(mods[n1]>tn)
       		rank1 = 0;
        else
   	    	rank1 = seeds[n1][tn+current];

       	if(mods[n2]>tn)
       		rank2 = 0;
        else
   	    	rank2 = seeds[n2][tn+current];

		return rank1 >= rank2;
		}
	else if(BACKOFFSCHEME == 2)
		{
	    for(int i=0; i<tn; i++)
   			{
       		mod1 /= 3;
        	mod2 /= 3;
   	    	}

	    if(mod1 < 1)
   			mod1 = 1;

    	if(mod2 < 1)
   			mod2 = 1;

	    return seeds[n1][tn+current]/mod1 >= seeds[n2][tn+current]/mod2;
		}
    }

int Seeds::whenIsMySlot(int tx, int rx)
	{
    bool now;

    for(int turn = current; turn < 2*numNodes; turn++)
    	{
        now = true;

        for(int j=0; j<numNodes; j++)
        	{
#ifndef FULLSEED
			// if a tx with higher rank than Tx is a neighbour of Rx
			//  or it's receiver is a neighbour Tx then defer.
         	if(j!=tx && j!=rx && isTx[j] && (adjMatrix[rx][j] ||
   	        	adjMatrix[tx][recv[j]] ) &&	first(j,tx,turn-current))
       	    	{
           		now = false;
               	break;
    			}
#else
			// if my Rx or a sender that is a neighbour of my Rx or a 2 hop
            //  neighbour that is sending
            //  to a 1 hop neighbour has a higher rank than me then defer.
			if(j!=tx && isTx[j] && (j==rx || adjMatrix[j][rx] || adjMatrix[tx][recv[j]])
            	&& first(j,tx,turn-current))
                {
                now = false;
                break;
                }
#endif
            }

        if(now)
        	return turn-current;
        }

    return 2*numNodes - current;
    }

